// start import
import TWTwainJS, {
  PdfDocumentSecurityLevel,
  PdfProtection,
  ScanSettings,
  ScannedFileSettings,
  TWConsole,
  TWDeviceName,
  TWDownloadMode,
  TWEvent,
  TWJSConnectionStatus,
  TWScanFinishedStatus,
  twConvertBinaryStringToUint8Array
} from "../src/TWTwainJS.js";

import { CapEnum, CapName } from "../src/TWTwainJSCapabilities.js";
// end import

// #region TWAIN functions
const TWStatus = {
  CONNECTING: "Connecting to TWTwain client",
  USER_ACTION: "Waiting for user action...",
  GET_DEVICE_INFO: "Retrieving device info",
  DEVICE_LIST: "Retrieving device list",
  DEVICE_LIST_ERROR: "Error retrieving device list",
  DEVICE_LIST_EMPTY: "No devices found",
  CURRENT_SETTINGS: "Retrieving current settings",
  SCAN: "Scanning"
};
let _currentPage = 0;
let _isWaiting = false;
let _deviceInfo = {}; // retrieve it from DEVICE_INFO_RETRIEVED event

let companyName = "TWDemoCompany";
let key = "TWDemoKey";
if (window.twTwainJSCompany) companyName = window.twTwainJSCompany;
if (window.twTwainJSKey) key = window.twTwainJSKey;
const twTwainJS = new TWTwainJS("twain1", companyName, key, twConsole);

//#endregion
// #region TWAIN JS Events
window.addEventListener(TWEvent.SOCKET_CONNECTION_STATUS_CHANGED, function (e) {
  twConsole(TWEvent.SOCKET_CONNECTION_STATUS_CHANGED + "[" + e.detail.twTwainJSInstance.name + "]", e.detail);
  if (e.detail.status === TWJSConnectionStatus.CONNECTED) {
    if (twTwainJS.info.clientVersion != twTwainJS.info.libraryVersion) {
      setStatus("", false);
      disableControls(true);
      return;
    }
    setStatus(TWStatus.DEVICE_LIST);
    setConnectionStatus(true);
    twTwainJS
      .getAvailableDevices()
      .then((result) => {
        console.log("getAvailableDevices promise:", result);
      })
      .catch((error) => {
        setError("getAvailableDevices error");
        console.log("getAvailableDevices promise error:", error);
        disableControls(true);
      });
  } else {
    if (e.detail.status === TWJSConnectionStatus.CLOSED) setError("Connection with the client closed");
    if (e.detail.status === TWJSConnectionStatus.ERROR) setError("Error in connection with the client");

    disableControls(true);
    setConnectionStatus(false);
  }
});

window.addEventListener(TWEvent.INIT_CONFIG, function (e) {
  twConsole(TWEvent.INIT_CONFIG + "[" + e.detail.twTwainJSInstance.name + "]", e.detail);
});

window.addEventListener(TWEvent.DEVICE_LIST_RETRIEVED, function (e) {
  twConsole(TWEvent.DEVICE_LIST_RETRIEVED + "[" + e.detail.twTwainJSInstance.name + "]", e.detail);
  const list = e.detail.devicesName;
  fillScanList(list);
  if (list.length === 0) {
    setStatus(TWStatus.DEVICE_LIST_EMPTY);
    return;
  }

  const cmbxScanners = document.getElementById("cmbxScanners");
  const defaultDeviceIndex = list.indexOf(e.detail.defaultDeviceName);
  if (defaultDeviceIndex > -1) {
    cmbxScanners.selectedIndex = defaultDeviceIndex;
    getDeviceInfo(e.detail.defaultDeviceName);
  } else setStatus(TWStatus.USER_ACTION, false);

  cmbxScanners.addEventListener("change", function () {
    getDeviceInfo(cmbxScanners.value);
  });
});

function getDeviceInfo(deviceName) {
  setStatus(TWStatus.GET_DEVICE_INFO);
  twTwainJS
    .getDeviceInfo(deviceName)
    .then((result) => {
      console.log("getDeviceInfo promise", result);
    })
    .catch((error) => {
      console.log("getDeviceInfo promise error", error);
    });
}

window.addEventListener(TWEvent.DEVICE_INFO_RETRIEVED, function (e) {
  twConsole(TWEvent.DEVICE_INFO_RETRIEVED + "[" + e.detail.twTwainJSInstance.name + "]", e.detail);
  _deviceInfo = e.detail.deviceInfo;
  const supportedCaps = _deviceInfo.supportedCapabilities;
  console.log(supportedCaps.getCapability(CapName.FEEDER_ENABLED), "color: green; background: yellow");
  console.log("%c" + _deviceInfo.currentSettings[CapName.FEEDER_ENABLED], "color: green; background: yellow");
  console.log("%c" + _deviceInfo.deviceName, "color: green; background: yellow");
  fillSelect("cmbxResolutions", supportedCaps.getCapability(CapName.X_RESOLUTION));
  fillSelect("cmbxPageSizes", supportedCaps.getCapability(CapName.SUPPORTED_SIZES));
  fillSelect("cmbxTransferTypes", supportedCaps.getCapability(CapName.IXFER_MECH));
  setSliderValues("sliderBrightness", supportedCaps.getCapability(CapName.BRIGHTNESS));
  setSliderValues("sliderThreshold", supportedCaps.getCapability(CapName.THRESHOLD));
  setSliderValues("sliderContrast", supportedCaps.getCapability(CapName.CONTRAST));
  setCurrentSettings(_deviceInfo.currentSettings);
  setStatus(TWStatus.USER_ACTION, false);

  /*setStatus(TWStatus.CURRENT_SETTINGS);
  const capNames = [
    CapName.SUPPORTED_SIZES,
    CapName.X_RESOLUTION,
    CapName.IXFER_MECH,
    CapName.DUPLEX_ENABLED,
    CapName.FEEDER_ENABLED,
    CapName.XFER_COUNT,
    CapName.THRESHOLD,
    CapName.CONTRAST,
    CapName.BRIGHTNESS,
    CapName.ROTATION
  ];
  twTwainJS
    .getCurrentSettings(capNames)
    .then((result) => {
      console.log("getCurrentSettings promise", result);
    })
    .catch((error) => {
      console.log("getCurrentSettings promise error", error);
    });
  */
});

window.addEventListener(TWEvent.CURRENT_SETTINGS_RECEIVED, function (e) {
  twConsole(TWEvent.CURRENT_SETTINGS_RECEIVED + "[" + e.detail.twTwainJSInstance.name + "]", e.detail);
  const settings = e.detail.currentSettings;
  setCurrentSettings(settings);
  setStatus(TWStatus.USER_ACTION, false);
});

function setCurrentSettings(settings) {
  setSliderValue("sliderThreshold", settings[CapName.THRESHOLD]);
  if (!_deviceInfo.supportedCapabilities.hasCapability(CapName.THRESHOLD))
    document.getElementById("sliderThreshold").disabled = true;

  setSliderValue("sliderBrightness", settings[CapName.BRIGHTNESS]);
  if (!_deviceInfo.supportedCapabilities.hasCapability(CapName.BRIGHTNESS))
    document.getElementById("sliderBrightness").disabled = true;

  setSliderValue("sliderContrast", settings[CapName.CONTRAST]);
  if (!_deviceInfo.supportedCapabilities.hasCapability(CapName.CONTRAST))
    document.getElementById("sliderContrast").disabled = true;

  setSelectValue("cmbxTransferTypes", settings[CapName.IXFER_MECH]);
  if (!_deviceInfo.supportedCapabilities.hasCapability(CapName.IXFER_MECH))
    document.getElementById("cmbxTransferTypes").disabled = true;

  setSelectValue("cmbxResolutions", settings[CapName.X_RESOLUTION]);
  if (!_deviceInfo.supportedCapabilities.hasCapability(CapName.X_RESOLUTION))
    document.getElementById("cmbxResolutions").disabled = true;

  setSelectValue("cmbxPageSizes", settings[CapName.SUPPORTED_SIZES]);
  if (!_deviceInfo.supportedCapabilities.hasCapability(CapName.SUPPORTED_SIZES))
    document.getElementById("cmbxPageSizes").disabled = true;

  switch (settings[CapName.IPIXEL_TYPE]) {
    case CapEnum.IPixelType.BW:
      document.getElementById("rbColorTypeBW").checked = true;
      break;
    case CapEnum.IPixelType.GRAY:
      document.getElementById("rbColorTypeGrayscale").checked = true;
      break;
    case CapEnum.IPixelType.RGB:
      document.getElementById("rbColorTypeColor").checked = true;
      break;
  }

  disableColorTypeSliders();
  document.getElementById("tbRotation").value = settings[CapName.ROTATION] == null ? 0 : settings[CapName.ROTATION];
  if (!_deviceInfo.supportedCapabilities.hasCapability(CapName.ROTATION))
    document.getElementById("tbRotation").disabled = true;

  document.getElementById("cbDuplex").checked = settings[CapName.DUPLEX_ENABLED];
  if (!_deviceInfo.supportedCapabilities.hasCapability(CapName.DUPLEX_ENABLED))
    document.getElementById("cbDuplex").disabled = true;

  document.getElementById("rbScanSourceAutofeed").checked = settings[CapName.FEEDER_ENABLED];
  document.getElementById("rbScanSourceFlatbed").checked = !settings[CapName.FEEDER_ENABLED];
  displayImageCountContainer();

  document.getElementById("tbImageCount").value =
    settings[CapName.XFER_COUNT] == null ? -1 : settings[CapName.XFER_COUNT];
  if (!_deviceInfo.supportedCapabilities.hasCapability(CapName.XFER_COUNT))
    document.getElementById("tbImageCount").disabled = true;
}

window.addEventListener(TWEvent.PAGE_SCANNED, function (e) {
  twConsole(TWEvent.PAGE_SCANNED + "[" + e.detail.twTwainJSInstance.name + "]", e.detail);
  _currentPage = e.detail.pageNumber;
  console.log(
    "Page scanned: " + _currentPage,
    "scanFormatType:",
    twTwainJS.scanResult.scanFormatType,
    "imageType:",
    twTwainJS.scanResult.imageType
  );
  displayPage(_currentPage);
});

window.addEventListener(TWEvent.SCAN_FINISHED, function (e) {
  twConsole(TWEvent.SCAN_FINISHED + "[" + e.detail.twTwainJSInstance.name + "]", e.detail);
  setStatus(TWStatus.USER_ACTION, false);
  if (e.detail.scanFinishedStatus == TWScanFinishedStatus.SUCCESS) {
    setInfo("Scan finished successfully");
  }
  if (e.detail.scanFinishedStatus == TWScanFinishedStatus.CANCELLED) {
    setInfo("Scan cancelled");
  }
  if (e.detail.scanFinishedStatus == TWScanFinishedStatus.ERROR) {
    setError("Scan finished with error." + e.detail.error);
  }
});

window.addEventListener(TWEvent.SINGLE_FILE_RETRIEVED, function (e) {
  twConsole(TWEvent.SINGLE_FILE_RETRIEVED + "[" + e.detail.twTwainJSInstance.name + "]", e.detail);
  setUploadStatus("File " + e.detail.fileName + " successfuly retrieved.", false, "info");
  uploadFileUsingFetch(e.detail.fileContent, e.detail.fileName);
});

window.addEventListener(TWEvent.FILE_RETRIEVE_FINISHED, function (e) {
  twConsole(TWEvent.FILE_RETRIEVE_FINISHED + "[" + e.detail.twTwainJSInstance.name + "]", e.detail);
});

window.addEventListener(TWEvent.SINGLE_FILE_DOWNLOADED, function (e) {
  twConsole(TWEvent.SINGLE_FILE_DOWNLOADED + "[" + e.detail.twTwainJSInstance.name + "]", e.detail);
  setDownloadStatus("File " + e.detail.fileName + " successfuly downloaded.", false, "info");
});

window.addEventListener(TWEvent.DOWNLOAD_FINISHED, function (e) {
  twConsole(TWEvent.DOWNLOAD_FINISHED + "[" + e.detail.twTwainJSInstance.name + "]", e.detail);
});

function getTWTwainClientDownloadUrl() {
  let clientDownloadUrl =
    "/downloads/twtwainjs/" +
    twTwainJS.info.libraryVersion +
    "/TWTwainClientSetup_" +
    twTwainJS.info.libraryVersion +
    ".exe";
  return clientDownloadUrl;
}

function getLatestVersionUrlText() {
  return (
    "<br/>The latest version is available <a href='" + getTWTwainClientDownloadUrl() + "' target='_blank'>here.</a>"
  );
}

function addDownloadUrlIfNeeded(error) {
  if (error.indexOf("No valid socket was found.") > -1) {
    error += "<br/>";
    error += "If the TWTwainClient.exe isn't installed yet, please install it.";
    error += getLatestVersionUrlText();
  }
  return error;
}

window.addEventListener(TWEvent.TW_ERROR, function (e) {
  twConsole(TWEvent.TW_ERROR + "[" + e.detail.twTwainJSInstance.name + "]", e.detail);
  let error = addDownloadUrlIfNeeded(e.detail.error);
  setStatus(TWStatus.USER_ACTION, false);
  setError(error);
});
// #endregion

// #region UI handling
document.addEventListener("DOMContentLoaded", function (event) {
  $(".toast").toast({ autohide: false });
  setStatus(TWStatus.CONNECTING);
  document.getElementById("twTwainJSVersion").innerText = twTwainJS.info.libraryVersion;

  twTwainJS
    .connectToClient(50000, 50025)
    .then((result) => {
      console.log("connectToClient promise:", result);
      if (result.status === TWJSConnectionStatus.CONNECTED) {
        document.getElementById("twTwainJSClientVersion").innerText = twTwainJS.info.clientVersion;
        if (twTwainJS.info.clientVersion != twTwainJS.info.libraryVersion) {
          let error =
            "The client version does not match the library version. Please update the client to ensure proper functionality.";
          error += getLatestVersionUrlText();
          setError(error);
        }
        console.log("connectToClient promise with status connected", result);
      }
    })
    .catch((error) => {
      console.log("connectToClient promise error:", error);
      error = addDownloadUrlIfNeeded(error);
      setStatus(TWStatus.USER_ACTION, false);
      setError(error);
      disableControls(true);
    });

  const scanBtn = document.getElementById("btnScan");
  scanBtn.addEventListener("click", function (evt) {
    scan();
  });

  const btnPrev = document.getElementById("btnPrevious");
  btnPrev.addEventListener("click", function (evt) {
    if (_currentPage > 1) {
      _currentPage--;
      displayPage(_currentPage);
    }
  });

  const btnNext = document.getElementById("btnNext");
  btnNext.addEventListener("click", function (evt) {
    if (_currentPage < twTwainJS.scanResult.pageCount()) {
      _currentPage++;
      displayPage(_currentPage);
    }
  });

  const accordionUploadBtnTxt = document.getElementById("accordion-btn-text");
  const collapseOne = document.getElementById("collapseOne");
  const accordionUploadIcon = document.getElementById("accordion-icon");

  accordionUploadBtnTxt.addEventListener("click", () => {
    if (collapseOne.classList.contains("show")) {
      accordionUploadBtnTxt.textContent = "Show upload status log details";
      accordionUploadIcon.classList.remove("fa-chevron-up");
      accordionUploadIcon.classList.add("fa-chevron-down");
    } else {
      accordionUploadBtnTxt.textContent = "Hide upload status log details";
      accordionUploadIcon.classList.remove("fa-chevron-down");
      accordionUploadIcon.classList.add("fa-chevron-up");
    }
  });

  const accordionBtn2 = document.getElementById("accordion-btn-text2");
  const collapseTwo = document.getElementById("collapseTwo");
  const accordionDownloadIcon = document.getElementById("accordion-icon2");

  accordionBtn2.addEventListener("click", () => {
    if (collapseTwo.classList.contains("show")) {
      accordionDownloadIcon.classList.remove("fa-chevron-up");
      accordionDownloadIcon.classList.add("fa-chevron-down");
      accordionBtn2.textContent = "Show download status log details";
    } else {
      accordionDownloadIcon.classList.remove("fa-chevron-down");
      accordionDownloadIcon.classList.add("fa-chevron-up");
      accordionBtn2.textContent = "Hide download status log details";
    }
  });

  const btnDownload = document.getElementById("btnDownload");
  btnDownload.addEventListener("click", function (evt) {
    setDownloadStatus("Preparing for download...");
    let downloadMode = document.getElementById("tbDownloadLocalLocation").checked
      ? TWDownloadMode.LOCAL
      : TWDownloadMode.STANDARD;
    let scannedFileSettings = new ScannedFileSettings();
    scannedFileSettings.fileName = document.getElementById("tbDownloadName").value;
    scannedFileSettings.pages = document.getElementById("tbPagesInputDownload").value;
    scannedFileSettings.isMultiPageFile = document.getElementById("cbMultiPageFile").checked;
    //it is possible to set also:
    // scannedFileSettings.tiffCompression
    // scannedFileSettings.pdfProtection
    // scannedFileSettings.imageQuality

    twTwainJS
      .downloadFile(downloadMode, scannedFileSettings)
      .then(function (response) {
        if (response.warning) setDownloadStatus("Download finished with warning." + response.warning, false, "warning");
        else setDownloadStatus("Download finished.", false, "success");
        console.log("in then...downloadFile", response);
      })
      .catch(function (error) {
        setDownloadStatus("Error in download.", false, "error");
        console.log("in catch...downloadFile", error);
      });
  });

  const btnUpload = document.getElementById("btnUpload");
  if (btnUpload) {
    btnUpload.addEventListener("click", function (evt) {
      let scannedFileSettings = new ScannedFileSettings();
      scannedFileSettings.fileName = document.getElementById("tbUploadName").value;
      scannedFileSettings.pages = document.getElementById("tbPagesInputUpload").value;
      scannedFileSettings.isMultiPageFile = document.getElementById("cbMultiPageFileUpload").checked;
      setUploadStatus("Retrieving file(s) for upload...");
      twTwainJS
        .getFile(scannedFileSettings)
        .then(function (response) {
          if (response.warning) {
            setUploadStatus("File(s) retrieve finished with warning. " + response.warning, false, "warning");
          } else {
            setUploadStatus("File(s) retrieve finished.", false, "info");
          }
        })
        .catch(function (error) {
          setUploadStatus("Error in upload.", false, "error");
          console.log("in catch...getFile", error);
        });
    });
  }

  const logSection = document.getElementById("logSection");
  const btnLog = document.getElementById("btnLog");
  btnLog.addEventListener("click", function (evt) {
    logSection.classList.toggle("d-none");
    if (logSection.classList.contains("d-none")) {
      btnLog.textContent = "Show Log";
    } else {
      btnLog.textContent = "Hide Log";
    }
  });

  const scanSourceRadioBtns = document.querySelectorAll('input[name="scanSource"]');
  scanSourceRadioBtns.forEach((radio) => {
    radio.addEventListener("click", displayImageCountContainer);
  });

  const cbPdfProtect = document.getElementById("cbPdfProtect");
  cbPdfProtect.addEventListener("change", function () {
    const controls = document.getElementsByClassName("twainJs__control3");
    for (let i = 0; i < controls.length; i++) {
      controls[i].disabled = !cbPdfProtect.checked;
    }
  });

  const colorTypeRadioBts = document.querySelectorAll('input[name="colorType"]');
  colorTypeRadioBts.forEach((radio) => {
    radio.addEventListener("click", disableColorTypeSliders);
  });

  const settingsRadioBtns = document.querySelectorAll('input[name="scanFormatType"]');
  settingsRadioBtns.forEach((radio) => {
    radio.addEventListener("click", () => {
      const imageQualityDiv = document.getElementById("imageQualityContainer");
      const protectPdfDiv = document.getElementById("protectPdfContainer");
      const tiffOptionsDiv = document.getElementById("tiffOptionsContainer");
      protectPdfDiv.classList.add("d-none");
      protectPdfDiv.classList.remove("d-flex");
      tiffOptionsDiv.classList.add("d-none");
      tiffOptionsDiv.classList.remove("d-flex");
      imageQualityDiv.classList.add("d-none");

      if (document.getElementById("rbScanFormatTypeJpeg").checked) {
        imageQualityDiv.classList.remove("d-none");
      }

      if (document.getElementById("rbScanFormatTypePdf").checked) {
        protectPdfDiv.classList.remove("d-none");
        protectPdfDiv.classList.add("d-flex");
        imageQualityDiv.classList.remove("d-none");
      }

      if (document.getElementById("rbScanFormatTypeTiff").checked) {
        tiffOptionsDiv.classList.remove("d-none");
        tiffOptionsDiv.classList.add("d-flex");
      }
    });
  });
});

function setInfo(message) {
  const info = document.getElementById("twStatusInfo");
  info.textContent = message;
  $(".toastInfo").toast("show");
}

function setError(message) {
  const error = document.getElementById("twStatusError");
  error.innerHTML = message;
  $(".toastError").toast("show");
}

function setConnectionStatus(connected) {
  const status = document.getElementById("twConnectionStatus");
  if (connected) {
    status.textContent = "Connected to client";
    status.classList.remove("badge-secondary");
    status.classList.add("badge-success");
  } else {
    status.textContent = "Client disconnected";
    status.classList.remove("badge-success");
    status.classList.add("badge-secondary");
  }
}

function setStatusClass(statusElem, statusType) {
  statusElem.classList.remove("text-success");
  statusElem.classList.remove("text-danger");
  statusElem.classList.remove("text-warning");

  if (statusType === "error") {
    statusElem.classList.add("text-danger");
  }

  if (statusType === "warning") {
    statusElem.classList.add("text-warning");
  }

  if (statusType === "success") {
    statusElem.classList.add("text-success");
  }
}

function setDownloadStatus(status, isWaiting = true, statusType = "info") {
  let statusElem = document.getElementById("twDownloadStatus");
  setStatusClass(statusElem, statusType);
  statusElem.textContent = status;
  let spinnerElem = document.getElementById("twDownloadStatusSpinner");
  if (!isWaiting) spinnerElem.classList.add("d-none");
  else spinnerElem.classList.remove("d-none");

  const controlElements = document.getElementsByClassName("twainJs__control2");
  for (let i = 0; i < controlElements.length; i++) {
    controlElements[i].disabled = isWaiting;
  }

  let statusClass = "";
  if (statusType === "error") {
    statusClass = "text-danger";
  }

  const downloadStatusList = document.getElementById("twDownloadStatusList");
  downloadStatusList.innerHTML += `<p class="mb-0 ${statusClass}">${status}</p>`;
}

function setUploadStatus(status, isWaiting = true, statusType = "info") {
  let statusElem = document.getElementById("twUploadStatus");
  setStatusClass(statusElem, statusType);
  statusElem.textContent = status;
  let spinnerElem = document.getElementById("twUploadStatusSpinner");
  if (!isWaiting) spinnerElem.classList.add("d-none");
  else spinnerElem.classList.remove("d-none");

  const controlElements = document.getElementsByClassName("twainJs__control3");
  for (let i = 0; i < controlElements.length; i++) {
    controlElements[i].disabled = isWaiting;
  }

  let statusClass = "";
  if (statusType === "error") {
    statusClass = "text-danger";
  }

  const uploadStatusList = document.getElementById("twUploadStatusList");
  uploadStatusList.innerHTML += `<p class="mb-0 ${statusClass}">${status}</p>`;
}

function setStatus(status, isWaiting = true) {
  let statusElem = document.getElementById("twStatus");
  statusElem.textContent = status;

  if (_isWaiting == isWaiting) {
    return;
  }

  _isWaiting = isWaiting;

  let spinnerElem = document.getElementById("twStatusSpinner");
  if (!isWaiting) spinnerElem.classList.add("d-none");
  else spinnerElem.classList.remove("d-none");

  disableControls(isWaiting);
}

function disableControls(disable) {
  const controlElements = document.getElementsByClassName("twainJs__control1");
  for (let i = 0; i < controlElements.length; i++) {
    controlElements[i].disabled = disable;
  }
}

function setSliderValues(id, cap) {
  const slider = document.getElementById(id);
  if (!cap || !cap.values) {
    slider.disabled = true;
    return;
  }

  slider.disabled = false;
  slider.min = cap.values[0];
  slider.max = cap.values[cap.values.length - 1];
}

function setSliderValue(id, value) {
  const slider = document.getElementById(id);
  slider.value = value;
}

function setSelectValue(selectName, value) {
  const select = document.getElementById(selectName);
  for (let i = 0; i < select.options.length; i++) {
    if (select.options[i].value == value) {
      select.selectedIndex = i;
      break;
    }
  }
}

function disableColorTypeSliders() {
  const rbColorTypeBW = document.getElementById("rbColorTypeBW");
  const sliderThreshold = document.getElementById("sliderThreshold");
  const sliderBrightness = document.getElementById("sliderBrightness");
  const sliderContrast = document.getElementById("sliderContrast");
  sliderThreshold.disabled = !rbColorTypeBW.checked;
  sliderBrightness.disabled = rbColorTypeBW.checked;
  sliderContrast.disabled = rbColorTypeBW.checked;
}

function displayImageCountContainer() {
  const imageCountContainer = document.getElementById("imageCountContainer");
  const rbScanSourceFlatbed = document.getElementById("rbScanSourceFlatbed");
  if (rbScanSourceFlatbed.checked) {
    imageCountContainer.classList.add("d-none");
    imageCountContainer.classList.remove("d-block");
  } else {
    imageCountContainer.classList.remove("d-none");
    imageCountContainer.classList.add("d-block");
  }
}

function displayPage(pageNumber) {
  const pageNum = document.getElementById("pageNum");
  pageNum.innerHTML = "Page " + pageNumber + "/" + twTwainJS.scanResult.pageCount();
  let div = document.getElementById("TWControlContainer");
  let canvas = div.querySelector("#twCanvas" + pageNumber);

  document.getElementById("btnUploadModal").disabled = false;
  document.getElementById("btnDownloadModal").disabled = false;
  document.getElementById("btnPrevious").disabled = pageNumber == 1;
  document.getElementById("btnNext").disabled = pageNumber == twTwainJS.scanResult.pageCount();

  let allCanvas = div.querySelectorAll("canvas");
  for (let i = 0; i < allCanvas.length; i++) {
    allCanvas[i].style.display = "none";
  }
  document.getElementById("twPlaceholder").style.display = "block";
  if (!canvas) {
    canvas = document.createElement("canvas");
    canvas.id = "twCanvas" + pageNumber;
    canvas.style.zIndex = 2;
    canvas.width = div.clientWidth;
    canvas.height = div.clientHeight;
    let ctx = canvas.getContext("2d");
    let image = new Image();
    image.onload = function () {
      let imageWidth = image.width;
      let imageHeight = image.height;
      if (imageWidth > canvas.width || imageHeight > canvas.height) {
        let xRatio = canvas.width / image.width;
        let yRatio = canvas.height / image.height;
        let ratio = Math.min(xRatio, yRatio);
        imageWidth = image.width * ratio;
        imageHeight = image.height * ratio;
      }
      let x = (canvas.width - imageWidth) / 2;
      let y = (canvas.height - imageHeight) / 2;
      ctx.drawImage(image, x, y, imageWidth, imageHeight);
    };
    image.src = twTwainJS.scanResult.getImageDataSrc(pageNumber);
    div.appendChild(canvas);
  }
  canvas.style.display = "block";
  document.getElementById("twPlaceholder").style.display = "none";
}

function fillScanList(values) {
  const select = document.getElementById("cmbxScanners");
  select.innerHTML = "";
  if (!values) {
    select.disabled = true;
    return;
  }
  select.disabled = false;

  for (let i = 0; i < values.length; i++) {
    const option = document.createElement("option");
    option.value = values[i];
    option.text = values[i];
    select.add(option);
  }
}

function fillSelect(id, cap) {
  const select = document.getElementById(id);
  select.innerHTML = "";
  if (!cap || !cap.values) {
    select.disabled = true;
    return;
  }
  select.disabled = false;

  for (let i = 0; i < cap.values.length; i++) {
    const option = document.createElement("option");
    option.value = cap.isEnum ? cap.values[i].value : cap.values[i];
    option.text = cap.isEnum ? cap.values[i].name : cap.values[i];
    if (
      // skip currently unsupported transfer types
      id == "cmbxTransferTypes" &&
      (option.value == CapEnum.IXferMech.FILE || option.value == CapEnum.IXferMech.MEMORY_FILE)
    )
      continue;
    select.add(option);
  }
}
// #endregion

// #region Log functionality
let logCounter = 0;
let logAscending = true;
let logStatusList = [];
document.addEventListener("DOMContentLoaded", function (event) {
  const btnClearMsgs = document.getElementById("btnClearMsgs");
  btnClearMsgs.addEventListener("click", function (evt) {
    let contentDivName = "debugContent";
    const contentDiv = document.getElementById(contentDivName);
    contentDiv.innerHTML = "";
    logStatusList = [];
  });

  const btnSort = document.getElementById("btnSort");
  btnSort.addEventListener("click", function (evt) {
    logAscending = logAscending ? false : true;
    btnSort.innerHTML = logAscending ? "Sort Descending" : "Sort Ascending";
    logStatusList.reverse();
    let contentDivName = "debugContent";
    const contentDiv = document.getElementById(contentDivName);
    contentDiv.innerHTML = "";
    logStatusList.forEach((element) => {
      contentDiv.appendChild(element);
    });
  });

  const debugContentContainer = document.getElementById("debugContent");
  debugContentContainer.addEventListener("click", function (event) {
    if (event.target.classList.contains("btn-link") || event.target.classList.contains("badge")) {
      const target = document.querySelector(event.target.getAttribute("data-target"));
      const targetId = target.getAttribute("id");
      const badge = document.querySelector(`[badge-target="${targetId}"]`);

      if (badge.innerHTML === "+") {
        badge.innerHTML = "-";
      } else {
        badge.innerHTML = "+";
      }
    }
  });
});

function twConsole(msg, msg2, type) {
  console.log(msg, msg2);
  addStatus(msg, msg2, type);
}

function jsonPropertyReplacer(key, value) {
  if (key.indexOf("_") === 0) return undefined;

  if (key == "fileContent") return "fileContent...";
  if (key == "images") {
    let images = [];
    for (let i = 0; i < value.length; i++) {
      images.push("Image" + i);
    }
    return images;
  }
  return value;
}

function addStatus(title, msgObj, type) {
  if (!title) return;
  let contentDivName = "debugContent";
  if (!type) type = TWConsole.EVENT;
  let json = "";
  if (type == TWConsole.SENT_MSG) {
    let jsonObj = JSON.parse(msgObj);
    json = JSON.stringify(jsonObj, jsonPropertyReplacer, 2);
  } else {
    json = JSON.stringify(msgObj, jsonPropertyReplacer, 2);
  }

  let statusId = "q" + logCounter++;
  const contentDiv = document.getElementById(contentDivName);

  const rowDiv = document.createElement("div");
  rowDiv.className = "row";
  const colDiv = document.createElement("div");
  colDiv.className = "col";
  const badgeLink = document.createElement("a");
  badgeLink.setAttribute("data-toggle", "collapse");
  badgeLink.setAttribute("data-target", "#" + statusId);
  badgeLink.setAttribute("aria-expanded", "false");
  badgeLink.setAttribute("aria-controls", statusId);
  badgeLink.setAttribute("badge-target", statusId);
  badgeLink.className = "badge badge-pill badge-primary twainJs__badge ";
  badgeLink.innerHTML = "+";

  const btn = document.createElement("button");
  btn.className = "btn btn-link " + type;
  btn.setAttribute("data-toggle", "collapse");
  btn.setAttribute("data-target", "#" + statusId);
  btn.setAttribute("aria-expanded", "false");
  btn.setAttribute("aria-controls", statusId);
  let typeTitle = "[TWEvent]";
  if (type == TWConsole.RECEIVED_MSG) typeTitle = "[TWReceivedMessage]";
  if (type == TWConsole.SENT_MSG) typeTitle = "[TWSentMessage]";
  btn.innerHTML = typeTitle + " " + title;

  const collapseDiv = document.createElement("div");
  collapseDiv.className = "collapse";
  collapseDiv.id = statusId;
  collapseDiv.setAttribute("aria-labelledby", statusId);

  const cardDiv = document.createElement("div");

  cardDiv.className = "card-body text-muted twainJs__status_scroll";
  const pre = document.createElement("pre");
  pre.innerText = json;
  cardDiv.appendChild(pre);

  collapseDiv.appendChild(cardDiv);
  colDiv.appendChild(badgeLink);
  colDiv.appendChild(btn);
  colDiv.appendChild(collapseDiv);
  rowDiv.appendChild(colDiv);
  if (logAscending) {
    contentDiv.appendChild(rowDiv);
  } else {
    contentDiv.prepend(rowDiv);
  }

  logStatusList.push(rowDiv);
}
// #endregion

function uploadFileUsingFetch(fileContent, fileName) {
  const apiUrl = document.getElementById("tbUploadApiUrl").value;
  let formData = new FormData();
  let content = twConvertBinaryStringToUint8Array(fileContent);
  const fileExtension = fileName.toLowerCase().split(".").pop();
  const fileType = fileExtension === "pdf" ? "application/pdf" : "image/" + fileExtension;
  const blob = new Blob(content, { type: fileType });
  let file = new File([blob], fileName, { type: fileType });

  formData.append("uploadFile", file);
  fetch(apiUrl, {
    method: "POST",
    body: formData
  })
    .then((response) => {
      if (!response.ok) {
        throw new Error("Network response was not ok");
      }
      return response.text();
    })
    .then((data) => {
      console.log(data);
      setUploadStatus("Successfully uploaded to " + data, false, "info");
    })
    .catch((error) => {
      console.log(error);
      let errorMessage = error.response ? error.response.data : error.message;
      setUploadStatus("Error in upload with fetch. " + errorMessage, false, "error");
    });
}

function scan() {
  // TW custom properties
  function setPDFProtection(scanFormatType, scanSettings) {
    if (scanFormatType != CapEnum.TWScanFormatType.PDF || !document.getElementById("cbPdfProtect").checked) {
      scanSettings.pdfProtection = null;
      return;
    }

    let p = new PdfProtection();
    p.ownerPassword = document.getElementById("tbOwnerPassword").value;
    p.userPassword = document.getElementById("tbUserPassword").value;
    p.permitAccessibilityExtractContent = document.getElementById("cbPermitAccessibilityExtractContent").checked;
    p.permitAnnotations = document.getElementById("cbPermitAnnotations").checked;
    p.permitAssembleDocument = document.getElementById("cbPermitAssembleDocument").checked;
    p.permitExtractContent = document.getElementById("cbPermitExtractContent").checked;
    p.permitFormsFill = document.getElementById("cbPermitFormsFill").checked;
    p.permitFullQualityPrint = document.getElementById("cbPermitFullQualityPrint").checked;
    p.permitPrint = document.getElementById("cbPermitPrint").checked;
    p.permitModifyDocument = document.getElementById("cbPermitModifyDocument").checked;
    p.securityLevel = document.getElementById("cmbxEncryption").value;

    scanSettings.pdfProtection = p;
  }

  function setTWCustomProperties(scanSettings) {
    scanSettings.showUI = document.getElementById("cbShowUI").checked;
    scanSettings.closeUIAfterAcquire = true;

    let scanFormatType = CapEnum.TWScanFormatType.BMP;
    if (document.getElementById("rbScanFormatTypeJpeg").checked) {
      scanFormatType = CapEnum.TWScanFormatType.JPEG;
    } else if (document.getElementById("rbScanFormatTypePng").checked) {
      scanFormatType = CapEnum.TWScanFormatType.PNG;
    } else if (document.getElementById("rbScanFormatTypeBmp").checked) {
      scanFormatType = CapEnum.TWScanFormatType.BMP;
    } else if (document.getElementById("rbScanFormatTypePdf").checked) {
      scanFormatType = CapEnum.TWScanFormatType.PDF;
    } else if (document.getElementById("rbScanFormatTypeTiff").checked) {
      scanFormatType = CapEnum.TWScanFormatType.TIFF;
    }

    scanSettings.tiffCompression = document.getElementById("cmbxTiffCompression").value;
    scanSettings.scanFormatType = scanFormatType;
    scanSettings.imageQuality = document.getElementById("tbImageQuality").value;

    setPDFProtection(scanFormatType, scanSettings);
  }

  function setTwainStandardProperties(scanSettings) {
    scanSettings.setCap(CapName.X_RESOLUTION, parseFloat(document.getElementById("cmbxResolutions").value));
    scanSettings.setCap(CapName.Y_RESOLUTION, parseFloat(document.getElementById("cmbxResolutions").value));

    scanSettings.setCap(CapName.FEEDER_ENABLED, document.getElementById("rbScanSourceAutofeed").checked);
    scanSettings.setCap(CapName.AUTO_FEED, document.getElementById("rbScanSourceAutofeed").checked);
    scanSettings.setCap(CapName.DUPLEX_ENABLED, document.getElementById("cbDuplex").checked);

    scanSettings.setCap(CapName.BRIGHTNESS, document.getElementById("sliderBrightness").value);
    scanSettings.setCap(CapName.THRESHOLD, document.getElementById("sliderThreshold").value);
    scanSettings.setCap(CapName.CONTRAST, document.getElementById("sliderContrast").value);

    scanSettings.setCap(CapName.ROTATION, document.getElementById("tbRotation").value);
    scanSettings.setCap(CapName.SUPPORTED_SIZES, document.getElementById("cmbxPageSizes").value);

    try {
      scanSettings.setCap(CapName.IXFER_MECH, parseInt(document.getElementById("cmbxTransferTypes").value));
    } catch (e) {
      console.log("setting IXFER_MECH failed: " + e);
    }

    let pixelType = CapEnum.IPixelType.BW;
    if (document.getElementById("rbColorTypeColor").checked) {
      pixelType = CapEnum.IPixelType.RGB;
    } else if (document.getElementById("rbColorTypeBW").checked) {
      pixelType = CapEnum.IPixelType.BW;
    } else if (document.getElementById("rbColorTypeGrayscale").checked) {
      pixelType = CapEnum.IPixelType.GRAY;
    }

    scanSettings.setCap(CapName.IPIXEL_TYPE, pixelType);
    scanSettings.setCap(CapName.XFER_COUNT, document.getElementById("tbImageCount").value);
  }

  let deviceName = TWDeviceName.DEFAULT;
  let supportedCapabilities;
  if (_deviceInfo) {
    deviceName = _deviceInfo.deviceName;
    supportedCapabilities = _deviceInfo.supportedCapabilities.capabilities;
  }

  let scanSettings = new ScanSettings(deviceName, supportedCapabilities);
  setTWCustomProperties(scanSettings);
  setTwainStandardProperties(scanSettings);

  let div = document.getElementById("TWControlContainer");
  let allCanvas = div.querySelectorAll("canvas");
  for (let i = 0; i < allCanvas.length; i++) {
    div.removeChild(allCanvas[i]);
  }

  document.getElementById("twPlaceholder").style.display = "initial";
  document.getElementById("pageNum").innerText = "Pages 0/0";
  document.getElementById("btnPrevious").disabled = true;
  document.getElementById("btnNext").disabled = true;
  document.getElementById("btnUploadModal").disabled = true;
  document.getElementById("btnDownloadModal").disabled = true;

  setStatus(TWStatus.SCAN);
  twTwainJS
    .scan(scanSettings)
    .then((result) => {
      console.log("scan result", result);
    })
    .catch((error) => {
      console.log("scan error", error);
    });
}
