// start import
import TWTwainJS, {
  ScanSettings,
  ScannedFileSettings,
  TWDownloadMode,
  TWEvent,
  TWJSConnectionStatus,
  TWScanFinishedStatus
} from "../src/TWTwainJS.js";

import { CapName } from "../src/TWTwainJSCapabilities.js";
// end import

let _isWaiting = true;
const waitStatus = "Waiting user action...";
let companyName = "TWDemoCompany";
let key = "TWDemoKey";
if (window.twTwainJSCompany) companyName = window.twTwainJSCompany;
if (window.twTwainJSKey) key = window.twTwainJSKey;
const twTwainJS = new TWTwainJS("twain1", companyName, key);
let fromPort = 50000; // you can define the port
let toPort = 50005;

document.addEventListener("DOMContentLoaded", function () {
  $(".toast").toast({ autohide: false });

  setStatus(waitStatus);
  document.getElementById("btnConnectToClient").addEventListener("click", function () {
    // creates socket connection by asking local Twain client to provide available port in
    // the range of fromPort-toPort
    setStatus("Connecting to the client");
    twTwainJS.connectToClient(fromPort, toPort).catch((error) => {
      setError("connectToClient error: " + error);
      console.log("connectToClient error", error);
    });
  });

  document.getElementById("btnGetScannerList").addEventListener("click", function () {
    setStatus("Retrieving the list of available scan devices");
    getAvailableDevices();
  });

  document.getElementById("btnGetDeviceInfo").addEventListener("click", function () {
    setStatus("Getting device info");
    getDeviceInfo(cmbxScanners.value);
  });

  document.getElementById("btnScan").addEventListener("click", function () {
    setStatus("Scanning");
    scan(cmbxScanners.value);
  });

  document.getElementById("btnDownloadFile").addEventListener("click", function () {
    setStatus("Downloading file");
    downloadFile();
  });
});

// every change of socket connection status triggers event TWEvent.SOCKET_CONNECTION_STATUS_CHANGED
// the most important one is information when then status is connected because in that moment
// it is possible to execute Twain commands on the local Twain client
window.addEventListener(TWEvent.SOCKET_CONNECTION_STATUS_CHANGED, function (e) {
  function setConnectionStatus(status) {
    let statusElem = document.getElementById("twConnectionStatus");
    statusElem.textContent = status;
  }

  console.log(TWEvent.SOCKET_CONNECTION_STATUS_CHANGED + "[" + e.detail.twTwainJSInstance.name + "]", e.detail);
  if (e.detail.status === TWJSConnectionStatus.CONNECTED) {
    setConnectionStatus(e.detail.status.name + " on port " + e.detail.twTwainJSInstance.port);
    btnConnectToClient.disabled = true;
    btnGetScannerList.disabled = false;
    // recommended place to execute some of your code like  twTwainJS.getAvailableDevices() or twTwainJS.scan(scanSettings)
    //getAvailableDevices();
  } else setConnectionStatus(e.detail.status.name);
  setStatus(waitStatus);
});

function setInfo(message) {
  const info = document.getElementById("twStatusInfo");
  info.textContent = message;
  $(".toastInfo").toast("show");
}

function setError(message) {
  const error = document.getElementById("twStatusError");
  error.textContent = message;
  $(".toastError").toast("show");
}

function setStatus(status, isWaiting = true, isError = false) {
  let statusElem = document.getElementById("twStatus");
  statusElem.textContent = status;

  if (_isWaiting == isWaiting) {
    return;
  }

  _isWaiting = isWaiting;

  let spinnerElem = document.getElementById("twStatusSpinner");
  if (!isWaiting) spinnerElem.classList.add("d-none");
  else spinnerElem.classList.remove("d-none");

  const controlElements = document.getElementsByClassName("twainJs__control1");
  for (let i = 0; i < controlElements.length; i++) {
    controlElements[i].disabled = isWaiting;
  }
}

function handleDownloadFileShowStatusMessage(message, show) {
  let statusMessageElem = document.getElementById("twDownloadFileStatus");
  let statusMessageWrapper = document.getElementById("twDownloadFileStatusWrapper");

  if (!show) statusMessageWrapper.classList.add("d-none");
  else {
    statusMessageWrapper.classList.remove("d-none");
    statusMessageElem.textContent = message;
  }
}

function handleScanShowStatusMessage(message, show) {
  let statusMessageElem = document.getElementById("twScanStatus");
  let statusMessageWrapper = document.getElementById("twScanStatusWrapper");

  if (!show) statusMessageWrapper.classList.add("d-none");
  else {
    statusMessageWrapper.classList.remove("d-none");
    statusMessageElem.textContent = message;
  }
}

function getAvailableDevices() {
  twTwainJS
    .getAvailableDevices()
    .then((result) => {
      console.log("Default device:", result.defaultDeviceName, "Devices:", result.devicesName);
      const values = result.devicesName;
      cmbxScanners.innerHTML = "";
      if (!values) {
        cmbxScanners.disabled = true;
        return;
      }
      cmbxScanners.disabled = false;

      for (let i = 0; i < values.length; i++) {
        const option = document.createElement("option");
        option.value = values[i];
        option.text = values[i];
        if (values[i] == result.defaultDeviceName) option.setAttribute("selected", "selected");
        cmbxScanners.add(option);
      }

      setStatus(waitStatus);
      btnGetDeviceInfo.disabled = values.length == 0;
    })
    .catch((error) => {
      setError("getAvailableDevices error: " + error);
      console.log("getAvailableDevices error", error);
    });
}

function showSupportedCapabilityProperties() {
  const capName = cmbxSupportedCapabilities.value;
  let cap = _deviceInfo.supportedCapabilities.getCapability(capName);

  console.log("Selected cap", cap);
  supportedCapabilityType.textContent = cap.type;
  cbIsMultiVal.checked = cap.isMultiVal;
  cbIsEnum.checked = cap.isEnum;
  cbIsReadOnly.checked = cap.isReadOnly;
  cmbxSupportedCapabilityValues.innerHTML = "";

  for (let i = 0; i < cap.values.length; i++) {
    const option = document.createElement("option");
    if (cap.isEnum) {
      option.value = cap.values[i].value;
      option.text = cap.values[i].name;
    } else {
      option.value = cap.values[i];
      option.text = cap.values[i];
    }
    if (i == 0) option.setAttribute("selected", "selected");
    cmbxSupportedCapabilityValues.add(option);
  }

  if (cap.isEnum) {
    let friendlyValue = cap.values.find((item) => item.value == _deviceInfo.currentSettings[capName]).name;
    currentSupportedCapabilityValue.textContent = friendlyValue;
  } else {
    currentSupportedCapabilityValue.textContent = _deviceInfo.currentSettings[capName];
  }

  if (cap.isEnum) {
    console.log("Supported values", cap.values);
  }
}

function showSettingsSupportedCapabilityProperties() {
  const capName = cmbxSettingsSupportedCapabilities.value;
  let cap = _deviceInfo.supportedCapabilities.getCapability(capName);

  cmbxSettingsSupportedCapabilityValues.innerHTML = "";
  cmbxSettingsMyCapabilityValue.innerHTML = "<option value='-100000'>Not set</option>";

  for (let i = 0; i < cap.values.length; i++) {
    const option = document.createElement("option");
    if (cap.isEnum) {
      option.value = cap.values[i].value;
      option.text = cap.values[i].name;
    } else {
      option.value = cap.values[i];
      option.text = cap.values[i];
    }
    if (i == 0) option.setAttribute("selected", "selected");
    cmbxSettingsSupportedCapabilityValues.add(option);
  }

  for (let i = 0; i < cap.values.length; i++) {
    const option = document.createElement("option");
    if (cap.isEnum) {
      option.value = cap.values[i].value;
      option.text = cap.values[i].name;
    } else {
      option.value = cap.values[i];
      option.text = cap.values[i];
    }
    if (option.value == cap.myValue) option.setAttribute("selected", "selected");
    cmbxSettingsMyCapabilityValue.add(option);
  }

  if (cap.isEnum) {
    let friendlyValue = cap.values.find((item) => item.value == _deviceInfo.currentSettings[capName]).name;
    currentSettingsSupportedCapabilityValue.textContent = friendlyValue;
  } else {
    currentSettingsSupportedCapabilityValue.textContent = _deviceInfo.currentSettings[capName];
  }

  if (cap.isEnum) {
    console.log("Supported values", cap.values);
  }
}

function fillSupportedCapabilities() {
  let supportedCapabilities = _deviceInfo.supportedCapabilities;
  let values = Object.keys(supportedCapabilities.capabilities);

  cmbxSupportedCapabilities.innerHTML = "";

  if (!values) {
    cmbxSupportedCapabilities.disabled = true;
    return;
  }

  for (let i = 0; i < values.length; i++) {
    const option = document.createElement("option");
    option.value = values[i];
    option.text = values[i];
    cmbxSupportedCapabilities.add(option);
  }
  cmbxSupportedCapabilities.disabled = false;
  cmbxSupportedCapabilities.addEventListener("change", showSupportedCapabilityProperties);

  showSupportedCapabilityProperties();
}

function fillSettingsCapabilities() {
  let supportedCapabilities = _deviceInfo.supportedCapabilities;
  let values = Object.keys(supportedCapabilities.capabilities);
  cmbxSettingsSupportedCapabilities.innerHTML = "";

  if (!values) {
    cmbxSettingsSupportedCapabilities.disabled = true;
    return;
  }

  cmbxSettingsSupportedCapabilities.disabled = false;

  cmbxSettingsSupportedCapabilities.addEventListener("change", showSettingsSupportedCapabilityProperties);

  cmbxSettingsMyCapabilityValue.addEventListener("change", function () {
    const capName = cmbxSettingsSupportedCapabilities.value;
    let cap = _deviceInfo.supportedCapabilities.getCapability(capName);
    let val = cmbxSettingsMyCapabilityValue.value;
    if (val == "-100000") cap.myValue = undefined;
    else cap.myValue = cmbxSettingsMyCapabilityValue.value;

    Array.from(cmbxSettingsSupportedCapabilities.options).find((option) => option.text == capName).style.fontWeight =
      cap.myValue === undefined ? "normal" : "bold";
  });

  for (let i = 0; i < values.length; i++) {
    const option = document.createElement("option");
    option.value = values[i];
    option.text = values[i];
    if (!supportedCapabilities.isReadOnly(values[i])) cmbxSettingsSupportedCapabilities.add(option);
  }

  showSettingsSupportedCapabilityProperties();
}

function fillDeviceInfoModal() {
  deviceName.textContent = _deviceInfo.deviceName;
  fillSupportedCapabilities();
  fillSettingsCapabilities();
}

let _deviceInfo;
function getDeviceInfo(deviceName) {
  //let deviceName = "your scanner device name";
  //deviceName = TWDeviceName.SELECT_FROM_UI; // shows UI to the client to select a device
  //deviceName = TWDeviceName.DEFAULT; // "" - it uses current client default device

  // if true, returned currentSettings will be empty.
  // Using that option can be useful if we don't care about currentSettings
  // and we want to have some performance gains (reading currentSettings isn't fast)
  const ignoreCurrentValues = false;
  twTwainJS
    .getDeviceInfo(deviceName, ignoreCurrentValues)
    .then((result) => {
      let deviceInfo = result.deviceInfo;
      btnSeeDeviceInfo.disabled = false;
      _deviceInfo = deviceInfo;
      console.log("Device info object", result.deviceInfo);
      console.log("Current settings", deviceInfo.currentSettings);

      fillDeviceInfoModal(deviceInfo);
      let supportedCapabilities = deviceInfo.supportedCapabilities;
      console.log("All Supported capabilities", supportedCapabilities.capabilities);

      // example how to check for some specific capability
      if (supportedCapabilities.hasCapability(CapName.X_RESOLUTION)) {
        const cap = supportedCapabilities.getCapability(CapName.X_RESOLUTION);
        console.log(
          `Supported values for ${CapName.X_RESOLUTION}:`,
          cap.values,
          "Is read only:",
          cap.isReadOnly,
          "Type:",
          cap.type,
          "IsEnum",
          cap.isEnum,
          "Is multi value",
          cap.isMultiVal
        );
      }
      // also are available methods:
      console.log("Is it read only?", supportedCapabilities.isReadOnly(CapName.X_RESOLUTION));
      console.log("Is supported value?", supportedCapabilities.isSupportedValue(CapName.X_RESOLUTION));
      btnShowScanModal.disabled = false;
      setStatus(waitStatus);
    })
    .catch((error) => {
      setError("getAvailableDevices error: " + error);
      console.log("getDeviceInfo error", error);
      setError("getDeviceInfo error " + error);
    });
}

function scan(deviceName) {
  //let deviceName = "your device name";
  //deviceName = TWDeviceName.DEFAULT; // it will use current default device
  // deviceName = TWDeviceName.SELECT_FROM_UI; // it will show UI to select one of available devices

  handleScanShowStatusMessage("", false);
  btnShowScanModal.disabled = true;
  let supportedCapabilities = {}; // optional. if empty there will not be checked if the setting is valid
  // option: to call before getDeviceInfo and use its supportedCapabilities
  // deviceName = deviceInfo.deviceName;
  supportedCapabilities = _deviceInfo.supportedCapabilities.capabilities;

  let scanSettings = new ScanSettings(deviceName, supportedCapabilities);
  Object.keys(_deviceInfo.supportedCapabilities.capabilities).forEach((capName) => {
    let cap = _deviceInfo.supportedCapabilities.getCapability(capName);
    if (cap.myValue) {
      scanSettings.setCap(capName, cap.myValue);
    }
  });
  // for tiff format it is possible to set tiffCompression
  scanSettings.tiffCompression = cmbxSettingsTiffCompression.value;

  // for pdf it is possible to set pdfProtection
  //scanSettings.pdfProtection = new PdfProtection();
  //scanSettings.pdfProtection.ownerPassword = "...";
  //scanSettings.pdfProtection... other properties like userPassword, permitPrint, permitExtractContent...

  // example how to set some specific capability
  //scanSettings.setCap(CapName.FEEDER_ENABLED, true);
  //scanSettings.setCap(CapName.SUPPORTED_SIZES, CapEnum.SupportedSizes.A4);
  // you can use any of avaialable standard TWAIN properties ->
  // CapName has a list of all available TWAIN property names

  //when setCap is called, if scanSettings.supportedCapabilities isn't empty it will check if the value
  //is valid for selected capability
  //it checks if capability is supported on that device and if it is read-only and if the value is supported

  scanSettings.showUI = settingsShowUI.checked;
  scanSettings.scanFormatType = cmbxSettingsScanFormatType.value;
  // it has effect only for pdf or jpg
  scanSettings.imageQuality = parseInt(settingsImageQuality.value);
  scanSettings.tiffCompression = cmbxSettingsTiffCompression.value;
  scanSettings.closeUIAfterAcquire = settingsCloseUIAfterAcquire.checked;
  // it triggers 2 events: TWEvent.PAGE_SCANNED and TWEvent.SCAN_FINISHED_SUCCESS
  twTwainJS
    .scan(scanSettings)
    .then((result) => {
      // promise is resolved when scan has finished
      console.log("scan result", result);
      // result.deviceName
      // result.twTwainJSInstance
      // twTwainJS.scanResult contains the result of the scan
      btnShowDownloadModal.disabled = false;
      if (result.scanFinishedStatus === TWScanFinishedStatus.SUCCESS) {
        handleScanShowStatusMessage(
          "Scan finished successfully. Total number of pages: " + twTwainJS.scanResult.pageCount(),
          true
        );
      }
      if (result.scanFinishedStatus === TWScanFinishedStatus.CANCELLED)
        handleScanShowStatusMessage("Scan cancelled.", true);
      if (result.scanFinishedStatus === TWScanFinishedStatus.ERROR) {
        handleScanShowStatusMessage("Scan finished with error.", true);
      }
    })
    .catch((error) => {
      btnShowScanModal.disabled = false;
      handleScanShowStatusMessage("Scan error: " + error, true);
      console.log("scan error", error);
    })
    .finally(() => {
      btnShowScanModal.disabled = false;
      setStatus(waitStatus);
    });
}

// event is triggered if you don't want to use promise for some reason
window.addEventListener(TWEvent.DEVICE_LIST_RETRIEVED, function (e) {
  console.log(TWEvent.DEVICE_LIST_RETRIEVED + "[" + e.detail.twTwainJSInstance.name + "]", e.detail);
  console.log("Default device:", e.detail.defaultDeviceName, "Devices:", e.detail.devicesName);
});

// event is triggered if you don't want to use promise for some reason
window.addEventListener(TWEvent.DEVICE_INFO_RETRIEVED, function (e) {
  console.log(TWEvent.DEVICE_INFO_RETRIEVED + "[" + e.detail.twTwainJSInstance.name + "]", e.detail);
  // e.detail.deviceInfo.supportedCapabilities
  // e.detail.deviceInfo.currentSettings
  // e.detail.deviceInfo.deviceName
});

// example how to display page in canvas when it is scanned
window.addEventListener(TWEvent.PAGE_SCANNED, function (e) {
  console.log(TWEvent.PAGE_SCANNED + "[" + e.detail.twTwainJSInstance.name + "]", e.detail);
  console.log("Current number of scanned pages:", twTwainJS.scanResult.pageCount());
  twScanStatus.textContent = "Current number of scanned pages:" + e.detail.pageNumber;
  // show this page in canvas
  /*let canvas = document.createElement("canvas");
  canvas.id = "twCanvas" + pageNumber;
  let ctx = canvas.getContext("2d");
  let image = new Image();
  image.onload = function () {
    canvas.width = image.width;
    canvas.height = image.height;
    ctx.drawImage(image, 0, 0, image.width, image.height);
  };
  image.src = twTwainJS.scanResult.getImageDataSrc(e.detail.pageNumber);
  document.body.appendChild(canvas);
  */
});

function downloadFile() {
  let scannedFileSettings = new ScannedFileSettings();
  scannedFileSettings.fileName = "scanned_file." + document.getElementById("cmbxSettingsDownloadFileType").value;
  scannedFileSettings.isMultiPageFile = document.getElementById("settingsMultiPage").checked;
  handleDownloadFileShowStatusMessage("", false);
  setStatus("Downloading file...");
  const downloadMode = document.getElementById("cbDownloadLocal").checked
    ? TWDownloadMode.LOCAL
    : TWDownloadMode.STANDARD;
  twTwainJS
    .downloadFile(downloadMode, scannedFileSettings)
    .then((result) => {
      console.log("file downloaded.", result);
      if (result.warning) {
        handleDownloadFileShowStatusMessage("Download finished with warning. " + result.warning, true);
      } else handleDownloadFileShowStatusMessage("Downloaded finished with success.", true);
      //result.pages
      //result.downloadMode -> it can be TWDownloadMode.STANDARD or TWDownloadMode.LOCAL
      //result.twTwainJSInstance
    })
    .catch((error) => {
      console.log("file download error", error);
      handleDownloadFileShowStatusMessage("File downloaded error:" + error, true);
    })
    .finally(() => {
      setStatus(waitStatus);
    });

  // this will create 3 jpg files (for pages 3, 2, 2) and standard browser download will be used
  // so visitor can choose its download location
  /* scannedFileSettings.isMultiPageFile = false;
  scannedFileSettings.fileName = "test.jpg";
  twTwainJS
    .downloadFile(TWDownloadMode.STANDARD, scannedFileSettings)
    .then((result) => {
      console.log("file downloaded.", result);
    })
    .catch((error) => {
      console.log("file download error", error);
    });
 */
}

window.addEventListener(TWEvent.SCAN_FINISHED_SUCCESS, function (e) {
  console.log(TWEvent.SCAN_FINISHED_SUCCESS + "[" + e.detail.twTwainJSInstance.name + "]", e.detail);
  // e.detail.scanFinishedStatus - it can be TWScanFinishedStatus.SUCCESS,
  // TWScanFinishedStatus.ERROR, TWScanFinishedStatus.CANCELLED
  // e.detail.twTwainJSInstance
});

window.addEventListener(TWEvent.SINGLE_FILE_DOWNLOADED, function (e) {
  console.log(TWEvent.SINGLE_FILE_DOWNLOADED + "[" + e.detail.twTwainJSInstance.name + "]", e.detail);
  //e.detail.fileName
  //e.detail.fileContent
  //e.detail.twTwainJSInstance
  //e.detail.downloadMode -> it can be TWDownloadMode.STANDARD or TWDownloadMode.LOCAL
});

// event is triggered if you don't want to use promise for some reason
window.addEventListener(TWEvent.DOWNLOAD_FINISHED, function (e) {
  console.log(TWEvent.DOWNLOAD_FINISHED + "[" + e.detail.twTwainJSInstance.name + "]", e.detail);
  //e.detail.warning
  //e.detail.pages
  //e.detail.twTwainJSInstance
  //e.detail.downloadMode -> it can be TWDownloadMode.STANDARD or TWDownloadMode.LOCAL
});

window.addEventListener(TWEvent.SINGLE_FILE_RETRIEVED, function (e) {
  console.log(TWEvent.SINGLE_FILE_RETRIEVED + "[" + e.detail.twTwainJSInstance.name + "]", e.detail);
  // e.detail.fileName
  // e.detail.fileContent
  // e.detail.twTwainJSInstance
});

// event TWEvent.FILE_RETRIEVE_FINISHED is trigger if you don't want to use a promise
window.addEventListener(TWEvent.FILE_RETRIEVE_FINISHED, function (e) {
  console.log(TWEvent.FILE_RETRIEVE_FINISHED + "[" + e.detail.twTwainJSInstance.name + "]", e.detail);
  // e.detail.warning
  // e.detail.twTwainJSInstance
});

//errors could be catched in two ways
//1. directly in .catch from the promise call
//2. by catching one of the available error events: TWEvent.SCAN_FINISHED_ERROR for scanning error or TWEvent.TW_ERROR for every other possible error
window.addEventListener(TWEvent.TW_ERROR, function (e) {
  console.log(TWEvent.TW_ERROR + "[" + e.detail.twTwainJSInstance.name + "]", e.detail); // e.detail.error
});
